@@ -296,7 +296,7 @@ int bch2_bkey_get_io_opts(struct btree_trans *trans,
 	if (!snapshot_opts) {
 		bch2_inode_opts_get(c, opts, metadata);
 
-		if (k.k->p.snapshot) {
+		if (!metadata && k.k->p.snapshot) {
 			struct bch_inode_unpacked inode;
 			int ret = bch2_inode_find_by_inum_snapshot(trans, k.k->p.inode, k.k->p.snapshot,
 								   &inode, BTREE_ITER_cached);
@@ -306,14 +306,16 @@ int bch2_bkey_get_io_opts(struct btree_trans *trans,
 				bch2_inode_opts_get_inode(c, &inode, opts);
 		}
 	} else {
-		if (snapshot_opts->fs_io_opts.change_cookie != atomic_read(&c->opt_change_cookie)) {
+		if (snapshot_opts->fs_io_opts.change_cookie	!= atomic_read(&c->opt_change_cookie) ||
+		    snapshot_opts->metadata			!= metadata) {
 			bch2_inode_opts_get(c, &snapshot_opts->fs_io_opts, metadata);
 
+			snapshot_opts->metadata = metadata;
 			snapshot_opts->cur_inum = 0;
 			snapshot_opts->d.nr = 0;
 		}
 
-		if (k.k->p.snapshot) {
+		if (!metadata && k.k->p.snapshot) {
 			if (snapshot_opts->cur_inum != k.k->p.inode) {
 				snapshot_opts->d.nr = 0;
 
@@ -362,6 +364,8 @@ int bch2_bkey_get_io_opts(struct btree_trans *trans,
 #undef x
 	}
 
+	BUG_ON(metadata && opts->erasure_code);
+
 	return 0;
 }
 
@@ -374,10 +378,46 @@ static const char * const bch2_rebalance_state_strs[] = {
 #undef x
 };
 
-int bch2_set_rebalance_needs_scan_trans(struct btree_trans *trans, u64 inum)
+static u64 rebalance_scan_encode(struct rebalance_scan s)
+{
+	switch (s.type) {
+	case REBALANCE_SCAN_fs:
+		return 0;
+	case REBALANCE_SCAN_metadata:
+		return 1;
+	case REBALANCE_SCAN_device:
+		return s.dev + 32;
+	case REBALANCE_SCAN_inum:
+		return s.inum;
+	default:
+		BUG();
+	}
+}
+
+static struct rebalance_scan rebalance_scan_decode(u64 v)
+{
+	if (v == 0)
+		return (struct rebalance_scan) { .type = REBALANCE_SCAN_fs };
+	if (v == 1)
+		return (struct rebalance_scan) { .type = REBALANCE_SCAN_metadata };
+	if (v < BCACHEFS_ROOT_INO)
+		return (struct rebalance_scan) {
+			.type = REBALANCE_SCAN_device,
+			.dev =  v - 32,
+	};
+
+	return (struct rebalance_scan) {
+		.type = REBALANCE_SCAN_inum,
+		.inum = v,
+	};
+}
+
+int bch2_set_rebalance_needs_scan_trans(struct btree_trans *trans, struct rebalance_scan s)
 {
 	CLASS(btree_iter, iter)(trans, BTREE_ID_rebalance_work,
-				SPOS(inum, REBALANCE_WORK_SCAN_OFFSET, U32_MAX),
+				SPOS(rebalance_scan_encode(s),
+				     REBALANCE_WORK_SCAN_OFFSET,
+				     U32_MAX),
 				BTREE_ITER_intent);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&iter));
 
@@ -394,16 +434,17 @@ int bch2_set_rebalance_needs_scan_trans(struct btree_trans *trans, u64 inum)
 	return bch2_trans_update(trans, &iter, &cookie->k_i, 0);
 }
 
-int bch2_set_rebalance_needs_scan(struct bch_fs *c, u64 inum)
+int bch2_set_rebalance_needs_scan(struct bch_fs *c, struct rebalance_scan s)
 {
 	CLASS(btree_trans, trans)(c);
 	return commit_do(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc,
-			    bch2_set_rebalance_needs_scan_trans(trans, inum));
+			    bch2_set_rebalance_needs_scan_trans(trans, s));
 }
 
 int bch2_set_fs_needs_rebalance(struct bch_fs *c)
 {
-	return bch2_set_rebalance_needs_scan(c, 0);
+	return bch2_set_rebalance_needs_scan(c,
+				(struct rebalance_scan) { .type = REBALANCE_SCAN_fs });
 }
 
 static int bch2_clear_rebalance_needs_scan(struct btree_trans *trans, u64 inum, u64 cookie)
@@ -491,11 +532,12 @@ static int rebalance_set_data_opts(struct btree_trans *trans,
 	struct bch_fs *c = trans->c;
 
 	memset(data_opts, 0, sizeof(*data_opts));
-	data_opts->rewrite_ptrs		= bch2_bkey_ptrs_need_rebalance(c, opts, k);
+	data_opts->type			= BCH_DATA_UPDATE_rebalance;
+	data_opts->ptrs_rewrite		= bch2_bkey_ptrs_need_rebalance(c, opts, k);
 	data_opts->target		= opts->background_target;
 	data_opts->write_flags		|= BCH_WRITE_only_specified_devs;
 
-	if (!data_opts->rewrite_ptrs) {
+	if (!data_opts->ptrs_rewrite) {
 		/*
 		 * device we would want to write to offline? devices in target
 		 * changed?
@@ -507,36 +549,6 @@ static int rebalance_set_data_opts(struct btree_trans *trans,
 		return 0;
 	}
 
-	if (trace_rebalance_extent_enabled()) {
-		CLASS(printbuf, buf)();
-
-		bch2_bkey_val_to_text(&buf, c, k);
-		prt_newline(&buf);
-
-		unsigned move_ptrs	= 0;
-		unsigned compress_ptrs	= 0;
-		u64 sectors		= 0;
-
-		bch2_bkey_needs_rebalance(c, k, opts, &move_ptrs, &compress_ptrs, &sectors);
-
-		if (move_ptrs) {
-			prt_str(&buf, "move=");
-			bch2_target_to_text(&buf, c, opts->background_target);
-			prt_str(&buf, " ");
-			bch2_prt_u64_base2(&buf, move_ptrs);
-			prt_newline(&buf);
-		}
-
-		if (compress_ptrs) {
-			prt_str(&buf, "compression=");
-			bch2_compression_opt_to_text(&buf, opts->background_compression);
-			prt_str(&buf, " ");
-			bch2_prt_u64_base2(&buf, compress_ptrs);
-			prt_newline(&buf);
-		}
-
-		trace_rebalance_extent(c, buf.buf);
-	}
 	count_event(c, rebalance_extent);
 	return 1;
 }
@@ -611,8 +623,44 @@ static int do_rebalance_scan_btree(struct moving_context *ctxt,
 	struct bch_fs *c = trans->c;
 	struct bch_fs_rebalance *r = &c->rebalance;
 
-	bch2_trans_begin(trans);
+	/*
+	 * peek(), peek_slot() don't know how to fetch btree root keys - we
+	 * really should fix this
+	 */
+	while (level == bch2_btree_id_root(c, btree)->level + 1) {
+		bch2_trans_begin(trans);
 
+		CLASS(btree_node_iter, iter)(trans, btree, start, 0, level - 1,
+					     BTREE_ITER_prefetch|
+					     BTREE_ITER_not_extents|
+					     BTREE_ITER_all_snapshots);
+		struct btree *b = bch2_btree_iter_peek_node(&iter);
+		int ret = PTR_ERR_OR_ZERO(b);
+		if (ret)
+			goto root_err;
+
+		if (b != btree_node_root(c, b))
+			continue;
+
+		if (btree_node_fake(b))
+			return 0;
+
+		struct bkey_s_c k = bkey_i_to_s_c(&b->key);
+
+		struct bch_inode_opts opts;
+		ret =   bch2_bkey_get_io_opts(trans, snapshot_io_opts, k, &opts) ?:
+			bch2_update_rebalance_opts(trans, &opts, &iter, k, SET_NEEDS_REBALANCE_opt_change);
+root_err:
+		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
+			continue;
+		if (bch2_err_matches(ret, BCH_ERR_data_update_fail))
+			ret = 0; /* failure for this extent, keep going */
+		WARN_ONCE(ret && !bch2_err_matches(ret, EROFS),
+			  "unhandled error from move_extent: %s", bch2_err_str(ret));
+		return ret;
+	}
+
+	bch2_trans_begin(trans);
 	CLASS(btree_node_iter, iter)(trans, btree, start, 0, level,
 				     BTREE_ITER_prefetch|
 				     BTREE_ITER_not_extents|
@@ -640,7 +688,7 @@ static int do_rebalance_scan_btree(struct moving_context *ctxt,
 noinline_for_stack
 static int do_rebalance_scan(struct moving_context *ctxt,
 			     struct per_snapshot_io_opts *snapshot_io_opts,
-			     u64 inum, u64 cookie, u64 *sectors_scanned)
+			     u64 scan_v, u64 cookie, u64 *sectors_scanned)
 {
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
@@ -651,7 +699,8 @@ static int do_rebalance_scan(struct moving_context *ctxt,
 
 	r->state = BCH_REBALANCE_scanning;
 
-	if (!inum) {
+	struct rebalance_scan s = rebalance_scan_decode(scan_v);
+	if (s.type == REBALANCE_SCAN_fs) {
 		r->scan_start	= BBPOS_MIN;
 		r->scan_end	= BBPOS_MAX;
 
@@ -663,16 +712,16 @@ static int do_rebalance_scan(struct moving_context *ctxt,
 			try(do_rebalance_scan_btree(ctxt, snapshot_io_opts, btree, 0,
 						    POS_MIN, SPOS_MAX));
 		}
-	} else {
-		r->scan_start	= BBPOS(BTREE_ID_extents, POS(inum, 0));
-		r->scan_end	= BBPOS(BTREE_ID_extents, POS(inum, U64_MAX));
+	} else if (s.type == REBALANCE_SCAN_inum) {
+		r->scan_start	= BBPOS(BTREE_ID_extents, POS(s.inum, 0));
+		r->scan_end	= BBPOS(BTREE_ID_extents, POS(s.inum, U64_MAX));
 
 		try(do_rebalance_scan_btree(ctxt, snapshot_io_opts, BTREE_ID_extents, 0,
 					    r->scan_start.pos, r->scan_end.pos));
 	}
 
 	try(commit_do(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc,
-			bch2_clear_rebalance_needs_scan(trans, inum, cookie)));
+			bch2_clear_rebalance_needs_scan(trans, scan_v, cookie)));
 
 	*sectors_scanned += atomic64_read(&r->scan_stats.sectors_seen);
 	/*
@@ -936,7 +985,7 @@ int bch2_fs_rebalance_init(struct bch_fs *c)
 static int check_rebalance_work_one(struct btree_trans *trans,
 				    struct btree_iter *extent_iter,
 				    struct btree_iter *rebalance_iter,
-				    struct bkey_buf *last_flushed)
+				    struct wb_maybe_flush *last_flushed)
 {
 	struct bch_fs *c = trans->c;
 	CLASS(printbuf, buf)();
@@ -949,7 +998,6 @@ static int check_rebalance_work_one(struct btree_trans *trans,
 	    extent_iter->btree_id == BTREE_ID_reflink &&
 	    (!rebalance_k.k ||
 	     rebalance_k.k->p.inode >= BCACHEFS_ROOT_INO)) {
-		bch2_trans_iter_exit(extent_iter);
 		bch2_trans_iter_init(trans, extent_iter,
 				     BTREE_ID_extents, POS_MIN,
 				     BTREE_ITER_prefetch|
@@ -1012,19 +1060,18 @@ int bch2_check_rebalance_work(struct bch_fs *c)
 	CLASS(btree_iter, rebalance_iter)(trans, BTREE_ID_rebalance_work, POS_MIN,
 					  BTREE_ITER_prefetch);
 
-	struct bkey_buf last_flushed __cleanup(bch2_bkey_buf_exit);
-	bch2_bkey_buf_init(&last_flushed);
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
 
 	struct progress_indicator_state progress;
 	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_rebalance_work));
 
 	int ret = 0;
-	while (!ret) {
-		progress_update_iter(trans, &progress, &rebalance_iter);
+	while (!(ret = lockrestart_do(trans,
+			progress_update_iter(trans, &progress, &rebalance_iter) ?:
+			wb_maybe_flush_inc(&last_flushed) ?:
+			check_rebalance_work_one(trans, &extent_iter, &rebalance_iter, &last_flushed))))
+	       ;
 
-		ret = lockrestart_do(trans,
-			check_rebalance_work_one(trans, &extent_iter, &rebalance_iter, &last_flushed));
-	}
-
-	return ret < 0 ? ret : 0;
+	return min(ret, 0);
 }
