@@ -913,6 +913,9 @@ static int check_inode(struct btree_trans *trans,
 	}
 
 	ret = bch2_check_inode_has_case_insensitive(trans, &u, &s->ids, &do_update);
+	if (bch2_err_matches(ret, ENOENT)) /* disconnected inode; will be fixed by a later pass */
+		ret = 0;
+	bch_err_msg(c, ret, "bch2_check_inode_has_case_insensitive()");
 	if (ret)
 		goto err;
 
@@ -1096,7 +1099,7 @@ int bch2_check_inodes(struct bch_fs *c)
 				POS_MIN,
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter);
+		progress_update_iter(trans, &progress, &iter) ?:
 		check_inode(trans, &iter, k, &snapshot_root, &s);
 	}));
 }
@@ -1187,7 +1190,7 @@ int bch2_check_unreachable_inodes(struct bch_fs *c)
 				POS_MIN,
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter);
+		progress_update_iter(trans, &progress, &iter) ?:
 		check_unreachable_inode(trans, &iter, k);
 	}));
 }
@@ -1627,7 +1630,7 @@ static int check_dirent(struct btree_trans *trans, struct btree_iter *iter,
 		new_d->k.p.inode	= d.k->p.inode;
 		new_d->k.p.snapshot	= d.k->p.snapshot;
 
-		struct btree_iter dup_iter = {};
+		CLASS(btree_iter_uninit, dup_iter)(trans);
 		return  bch2_hash_delete_at(trans,
 					    bch2_dirent_hash_desc, hash_info, iter,
 					    BTREE_UPDATE_internal_snapshot_node) ?:
@@ -1715,7 +1718,7 @@ again:
 				POS(BCACHEFS_ROOT_INO, 0),
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-			progress_update_iter(trans, &progress, &iter);
+			progress_update_iter(trans, &progress, &iter) ?:
 			check_dirent(trans, &iter, k, &hash_info, &dir, &target, &s,
 				     &need_second_pass);
 		})) ?:
@@ -1782,7 +1785,7 @@ int bch2_check_xattrs(struct bch_fs *c)
 			k,
 			NULL, NULL,
 			BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter);
+		progress_update_iter(trans, &progress, &iter) ?:
 		check_xattr(trans, &iter, k, &hash_info, &inode);
 	}));
 	return ret;
